"""Regularization proof of concept."""
import matplotlib.pyplot as plt
import numpy as np


def set_up_point_matrix(axis_x: np.ndarray, degree: int) -> np.ndarray:
    """Set up a point matrix to fit a polynomial.

    The matrix should have to following form:
    [a_1**0       a_1**1      ...  a_1**(degree-1)
     a_2**0       a_2**1      ...  a_2**(degree-1)
     a_3**0       a_3**1      ...  a_3**(degree-1)
     ...          ...         ...  ...
     a_points**0  a_points**1 ...  a_points**(degree-1)]

    Where the entries in the matrix are from the axis_x vector.

    Args:
        axis_x (np.ndarray): The values of the time or x-axis.
        degree (int): The degree of the polynomial.

    Returns:
        np.ndarray: The polynomial point matrix A.
    """
    mat_a = np.zeros((axis_x.shape[0], degree))
    for i in range(0, degree):
        mat_a[:, i] = axis_x**i
    return mat_a


if __name__ == "__main__":
    np.random.seed(8)
    b = np.array([0.0, 0.5, 1.0, 0.75, 1.5, 2.0, 0.0])

    x_axis = np.linspace(0, 1, num=len(b))
    A1 = set_up_point_matrix(x_axis, len(b))
    est_x = np.linalg.pinv(A1) @ b

    point_no = 300
    x_axis2 = np.linspace(0, 1, num=point_no)
    A2 = set_up_point_matrix(x_axis2, len(b))

    alpha = 0.2
    b_true = A2 @ np.linalg.pinv(A1) @ b
    b_noise = b_true + alpha * np.random.randn(point_no)

    plt.plot(x_axis2, b_true, label="b_true")
    plt.plot(x_axis2, b_noise, ".", label="b_noise")
    plt.legend()
    plt.show()

    # Regression
    A_lin = set_up_point_matrix(x_axis2, degree=2)
    x = np.linalg.pinv(A_lin) @ b_noise
    plt.plot(x_axis2, b_noise, ".", label="b_noise")
    plt.plot(x_axis2, x[0] + x[1] * x_axis2)
    plt.show()

    # Higher order Polynomial
    A3 = set_up_point_matrix(x_axis2, degree=300)
    x_poly = np.linalg.pinv(A3) @ b_noise
    plt.plot(x_axis2, b_noise, ".", label="b_noise")
    plt.plot(x_axis2, A3 @ x_poly)
    plt.show()

    # regularize to find the noise free solution.
    U, sigma, V = np.linalg.svd(A3)
    plt.plot(x_axis2, b_noise, ".")

    for scale in [1e-1, 1e-6, 1e-12]:  # , 0.]:
        # x = np.zeros(point_no)
        # for i in range(point_no):
        #     f = sigma[i]**2 / (sigma[i]**2 + scale)
        #     x = x + f*((U[:,i].T @ b_noise)/sigma[i]) * V.T[:,i]
        f_vec = sigma**2 / (sigma**2 + scale)
        F = np.diag(f_vec)
        # x2 = np.sum(V.T @ np.diag(U.T @ b_noise / sigma)@F, 1)
        x2 = V.T @ F @ np.diag(sigma ** (-1)) @ U.T @ b_noise

        plt.plot(x_axis2, A3 @ x2, label=str(scale))
    plt.legend()
    plt.ylim(-1.0, 3.0)
    plt.grid()
    plt.show()

    # Reducing the model complexity
    for degree in range(1, 20):
        A_matrix = set_up_point_matrix(x_axis2, degree)
        x_poly = np.linalg.pinv(A_matrix) @ b_noise
        result = A_matrix @ x_poly
        mse_error = ((result - b_noise) ** 2).mean()
        plt.scatter(degree, mse_error, c="b")
    plt.xlabel("Degree of polynomial")
    plt.ylabel("Mean Squared Error")
    plt.xticks(range(0, 20, 2))
    plt.grid("equal")
    plt.show()

    A_matrix = set_up_point_matrix(x_axis2, 7)
    x_poly = np.linalg.pinv(A_matrix) @ b_noise
    plt.plot(x_axis2, b_noise, ".", label="b_noise")
    plt.plot(x_axis2, A_matrix @ x_poly)
    plt.show()
